import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from active_bo_ros.PolicyModel.GaussianModelMultiDim import GaussianRBF

from active_bo_ros.AcquisitionFunctions.ExpectedImprovement import ExpectedImprovement
from active_bo_ros.AcquisitionFunctions.ProbabilityOfImprovement import ProbabilityOfImprovement
from active_bo_ros.AcquisitionFunctions.ConfidenceBound import ConfidenceBound
from active_bo_ros.AcquisitionFunctions.PreferenceExpectedImprovement import PreferenceExpectedImprovement

from sklearn.exceptions import ConvergenceWarning
import warnings

warnings.filterwarnings('ignore', category=ConvergenceWarning)


class BayesianOptimization:
    def __init__(self, nr_steps, nr_dims, nr_weights, acq='ei', seed=None,
                 fixed_dims=None, lower_bound=None, upper_bound=None):
        self.acq = acq
        self.episode = 0

        self.nr_steps = nr_steps
        self.nr_dims = nr_dims
        self.nr_weights = nr_weights
        self.weights = self.nr_weights * self.nr_dims

        if lower_bound is None:
            self.lower_bound = [-1.]*nr_dims
            self.lower_bound_1D = [-1.] * self.weights
        else:
            self.lower_bound = lower_bound
            self.lower_bound_1D = [x for x in lower_bound for _ in range(self.nr_weights)]

        if upper_bound is None:
            self.upper_bound = [1.]*nr_dims
            self.upper_bound_1D = [1.] * self.weights
        else:
            self.upper_bound = upper_bound
            self.upper_bound_1D = [x for x in upper_bound for _ in range(self.nr_weights)]

        self.seed = seed

        self.X = None
        self.Y = None
        self.gp = None

        self.fixed_dims = fixed_dims

        if fixed_dims is not None:
            self.fixed_dims_1D = self.convert_fixed_dims(fixed_dims, nr_weights)

        self.policy_model = GaussianRBF(self.nr_steps, self.nr_weights, self.nr_dims,
                                        lower_bound=self.lower_bound, upper_bound=self.upper_bound, seed=seed,
                                        fixed_dims=fixed_dims)

        self.acq_sample_size = 100

        self.best_reward = np.empty((1, 1))

        if acq == "Preference Expected Improvement":
            self.acq_fun = PreferenceExpectedImprovement(self.weights,
                                                         self.acq_sample_size,
                                                         initial_variance=10.0,
                                                         update_variance=0.05,
                                                         seed=seed,
                                                         fixed_dims=self.fixed_dims_1D,
                                                         lower_bound=self.lower_bound_1D,
                                                         upper_bound=self.upper_bound_1D,)

        self.reset_bo()

    @staticmethod
    def convert_fixed_dims(fixed_dims, nr_weights):
        new_fixed_dims = {}
        for dim, values in fixed_dims.items():
            for idx, value in enumerate(values):
                new_dim = dim + idx * nr_weights
                new_fixed_dims[new_dim] = value
        return new_fixed_dims

    def reset_bo(self):
        self.gp = GaussianProcessRegressor(Matern(nu=1.5, ), n_restarts_optimizer=5)  # length_scale=(1e-8, 1e5)
        self.best_reward = np.empty((1, 1))
        self.X = np.zeros((1, self.weights), dtype=np.float64)
        self.Y = np.zeros((1, 1), dtype=np.float64)
        self.episode = 0

    def next_observation(self):
        if self.acq == "Expected Improvement":
            x_next = ExpectedImprovement(self.gp,
                                         self.X,
                                         self.acq_sample_size,
                                         self.weights,
                                         kappa=0,
                                         seed=self.seed,
                                         lower=self.lower_bound,
                                         upper=self.upper_bound)

        elif self.acq == "Probability of Improvement":
            x_next = ProbabilityOfImprovement(self.gp,
                                              self.X,
                                              self.acq_sample_size,
                                              self.weights,
                                              kappa=0,
                                              seed=self.seed,
                                              lower=self.lower_bound,
                                              upper=self.upper_bound)

        elif self.acq == "Upper Confidence Bound":
            x_next = ConfidenceBound(self.gp,
                                     self.acq_sample_size,
                                     self.weights,
                                     beta=2.576,
                                     seed=self.seed,
                                     lower=self.lower_bound,
                                     upper=self.upper_bound)

        elif self.acq == "Preference Expected Improvement":
            x_next = self.acq_fun.expected_improvement(self.gp,
                                                       self.X,
                                                       kappa=0)

        else:
            raise NotImplementedError

        return x_next

    def add_observation(self, reward, x):
        if self.episode == 0:
            self.X[0, :] = x
            self.Y[0] = reward
            self.best_reward[0] = np.max(self.Y)
        else:
            self.X = np.vstack((self.X, np.around(x, decimals=8)))
            self.Y = np.vstack((self.Y, reward))
            self.best_reward = np.vstack((self.best_reward, np.max(self.Y)))

        self.gp.fit(self.X, self.Y)
        self.episode += 1

    def get_best_result(self):
        y_max = np.max(self.Y)
        idx = np.argmax(self.Y)
        x_max = self.X[idx, :].reshape((self.nr_weights, self.nr_dims), order='F')

        return y_max, x_max, idx
