
import numpy as np


class GaussianRBF:
    def __init__(self, nr_steps, nr_weights, nr_dims, lower_bound=None, upper_bound=None, seed=None, fixed_dims=None):
        self.nr_weights = nr_weights
        self.nr_steps = nr_steps
        self.nr_dims = nr_dims

        self.weights = None
        self.trajectory = None

        if lower_bound is None:
            self.lower_bound = [-1.]*nr_dims
        else:
            self.lower_bound = lower_bound

        if upper_bound is None:
            self.upper_bound = [1.]*nr_dims
        else:
            self.upper_bound = upper_bound

        self.rng = np.random.default_rng(seed=seed)

        self.fixed_dims = fixed_dims

        # initialize
        self.mid_points = np.linspace(0, self.nr_steps, self.nr_weights)
        if nr_weights > 1:
            self.std = self.mid_points[1] / (2 * np.sqrt(2 * np.log(2)))  # Full width at half maximum
        else:
            self.std = self.nr_steps / 2

        self.reset()

    def reset(self):
        self.weights = np.zeros((self.nr_weights, self.nr_dims))
        self.trajectory = np.zeros((self.nr_steps, self.nr_dims))

    def random_weights(self):
        for w in range(self.nr_weights):
            # If dim exists in fixed_dims, set weights directly
            if w in self.fixed_dims:
                self.weights[w, :] = np.array(self.fixed_dims[w])
            else:
                for dim in range(self.nr_dims):
                    self.weights[w, dim] = self.rng.uniform(self.lower_bound[dim], self.upper_bound[dim], 1)
        return self.weights

    def rollout(self):
        self.trajectory = np.zeros((self.nr_steps, self.nr_dims))
        for step in range(self.nr_steps):
            for weight in range(self.nr_weights):
                base_fun = np.exp(-0.5 * (step - self.mid_points[weight]) ** 2 / self.std ** 2)
                for dim in range(self.nr_dims):
                    self.trajectory[step, dim] += base_fun * self.weights[weight, dim]

        return self.trajectory

    def set_weights(self, x):
        self.weights = x.reshape((self.nr_weights, self.nr_dims), order='F')

    def get_x(self):
        return self.weights.flatten('F')
